/*
 * $LastChangedBy$
 * $LastChangedDate$
 * $LastChangedRevision$
 * $HeadURL$
 *
 * Copyright 2014 BlueConic Inc./BlueConic B.V. All rights reserved.
 */

import Foundation
import UIKit
/**
 Implementation of the BlueConic client, handling the profile retrieval and storage.
 This may be from cache, persistent storage on the client or direct requests to the BlueConic server.

 &lt;pre&gt:
 // Swift:
 import BlueConicClient

 // Objective-C:
 #import &lt;BlueConicClient/BlueConicClient-Swift.h&gt;
 &lt;/pre&gt:

 */
class BlueConicClientImpl {
    // List of private values
    // CommitLog, Cache and Reachability are part of the connecting architecture.
    private var _commitLog: BlueConicCommitLog!
    private var _requestCommitLog: BlueConicCommitLog!
    private var _cache: BlueConicCache!
    
    private var _isEnabled: Bool = true

    // Profile properties
    private var _profileId: String?
    private var _zoneId: String?

    // Configuration properties
    private var _operationQueue: OperationQueue!
    private var _updateRequired: Bool = false

    // Simulator properties
    private var _sessionId: String?
    private var _userName: String?

    // Plugin properties
    // Plugin class became @deprecated use BlueConicPlugin
    private var _currentPlugins = [AnyObject]()
    private var _plugins = [String: AnyClass]()

    // Connection properties
    private var _connections = [Connection]()
    
    // Segment properties
    private var _segments = [BlueConicSegment]()

    // Session time measure
    private var _sessionStartTime: Double = 0.0
    private var _sessionStopTime: Double = 0.0
    private var _activeContexts: Int = 0

    // Current context property
    private weak var _context: UIViewController?
    private var _locale: String?
    private var _screenName: String = ""
    private final var _configuration: BlueConicConfiguration!
    
    // Sharing
    private var _session: URLSession?

    /**
     Private constructor
     Creates a new BlueConicClient instance. This instance loads the cache and commitlog state
     from disk and starts task for syncing state and checking network state.

     - parameter context: The current view controlller
     */
    convenience init(context: UIViewController?) {
        self.init()
        self.setContext(context)
    }

    /**
     Constructor
     */
    init() {
        let startTime = CFAbsoluteTimeGetCurrent()
        // Make sure unsaved profile properties will be sent to the server.
        self._updateRequired = true
        self._configuration = BlueConicConfiguration()
        
        // Get the debug status from the app plist
        BlueConicDataManager.cleanUpFiles()
        
        // Get the Enabled/Disabled status
        self._isEnabled = self.loadEnabled()
        
        BCLog("Hostname: \(self._configuration.getHostName())")
        BCLog("App ID: \(BlueConicConfiguration.getPackageName())")

        // Get the commitlog from disk if present
        self._commitLog = self.loadCommitLog(Constants.Files.COMMITLOG)
        self._requestCommitLog = self.loadCommitLog(Constants.Files.REQUEST_COMMITLOG)

        // Get the cache from disk if present
        self._cache = self.loadCache()

        // Load the profile id from the previous session
        self._profileId = self.loadProfileId()

        // Create an operation queue with one thread
        self._operationQueue = OperationQueue()
        self._operationQueue.maxConcurrentOperationCount = 1
        
        self.registerPluginClasses()

        BCLog("Create new BlueConic client in (\(String(format:"%.2f", ((CFAbsoluteTimeGetCurrent() as Double) - startTime) * 1000)) ms)")
        BCLog("BlueConic client is \(self._isEnabled ? "enabled": "disabled")")
    }

    /**
     Returns the first value for a given profile property.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     let hobby: String = client.getProfileValue(&quot;hobby&quot;)

     // Objective-C:
     NSString* hobby = [client getProfileValue:@&quot;hobby&quot;];
     &lt;/pre&gt:

     - parameter property: The profile property to get the values for.

     - returns: The first value
     */
    internal func getProfileValue(_ property: String) -> String {
        return self._cache.getProfileValue(property)
    }

    /**
     Return the values for a given profile property.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     let hobbies: [String] = client.getProfileValues(&quot;hobbies&quot;)

     // Objective-C:
     NSArray* hobbies = [client getProfileValues:@&quot;hobbies&quot;];
     &lt;/pre&gt:

     - parameter property: The profile property to get the values for.

     - returns: A collection containing the values.
     */
    internal func getProfileValues(_ property: String) -> [String]? {
        return self._cache.getProfileValues(property)
    }

    /**
     Returns the current ViewController.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     let viewController = client.getViewController()

     // Objective-C:
     UIViewController* viewController = [client getViewController];
     &lt;/pre&gt:

     - returns:     The current ViewController.
     */
    internal func getViewController() -> UIViewController? {
        return self.getContext()
    }

    /**
     Returns a view component based on the given identifier or <code>nil</code> is no match is found.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     @IBOutlet weak var view: UIView!
     let view: UIView? = client.getView(&quot;#view&quot;)

     // Objective-C:
     @property (weak, nonatomic) IBOutlet UIView* view;
     UIView* view = [client getView:@&quot;#view&quot;];
     &lt;/pre&gt:

     - parameter selector: The Identifier, e.g. &quot;#view&quot;.

     - returns: The view or <code>nil</code>
     */
    internal func getView(_ selector: String) -> UIView? {
        return BlueConicClientImpl.getView(self.getContext(), selector: selector)
    }

    /**
     Returns the screenName either set in createEvent or the ViewControllers title.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     client.createEvent(&quot;PAGEVIEW&quot;, properties: [&quot;screenName&quot;: &quot;Main/HOMETAB&quot;])
     var screenName: String = client.getScreenName()

     // Objective-C:
     [client createEvent:@&quot;PAGEVIEW&quot; properties:@{@&quot;screenName&quot;: @&quot;MAIN/HOMETAB&quot;}];
     NSString* screenName = [client getScreenName];
     &lt;/pre&gt:

     - returns: The screen name
     */
    internal func getScreenName() -> String {
        return self._screenName
    }

    /**
     Return the ID of the profile

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     var profileId: String = client.getProfileId()

     // Objective-C:
     NSString* profileId = [client getProfileId];
     &lt;/pre&gt:

     - returns: The ID of the profile
     */
    internal func getProfileId() -> String? {
        return self._profileId
    }
    
    /**
     Is enabled

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     var isEnabled: Bool = client.isEnabled()

     // Objective-C:
     NSString* profileId = [client getProfileId];
     &lt;/pre&gt:

     - returns: The ID of the profile
     */
    internal func isEnabled() -> Bool {
        return self._isEnabled
    }

    /**
     Returns the screen name. The screen name is determined in the following way:<br />
     1. If the {@code properties} hold an entry 'screenName' that is used.<br />
     2. If the {@code properties} hold an entry 'location' that is used (backwards compatible).<br />
     - parameter properties: The properties.
     - returns: The screen name from the properties.
     */
    func getScreenNameFromProperties(_ properties: [String: Any]?) -> String? {
        var screenName: String? = nil
        if let properties = properties {
            var overrule = properties["screenName"] as? String

            if overrule == nil || overrule == "" {
                overrule = properties["location"] as? String
            }

            if let overrule = overrule {
                screenName = overrule.hasPrefix("/") ? String(overrule.dropFirst()) : overrule
            } else if overrule == nil && self.getContext() != nil {
                // There is no Screen name set in the properties, fallback to the title of the page.
                screenName = self.getContext()?.title ?? ""
                BCLog("There is no screen name set to the properties! Using the title '\(screenName ?? "")' instead.", flag: Constants.Flag.WARN)
            }
        }
        return screenName
    }

    /**
     Returns a component for the interaction.

     - parameter context: The current viewcontroller.
     - parameter selector: The id the of the component for the interaction.
     - returns: The component matching the selector or the position of the interaction or null if no match is found.
     */
    class func getView(_ context: UIViewController?, selector: String?) -> UIView? {
        var view: UIView?
        if let selector = selector, selector.hasPrefix(Constants.ID_PREFIX) {
            let id = String(selector.dropFirst())
            if let controller = context, controller.responds(to: Selector(id)) {
                view = controller.value(forKey: id) as? UIView
            }
        }

        return view
    }

    /**
     Adds a single property value to the profile.
     If there are already values for a property the new value will be added.
     Values for a property need to be unique; passing the same value multiple times will have no effect.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     client.addProfileValue(&quot;hobbies&quot;, value:&quot;tennis&quot;)

     // Objective-C:
     [client addProfileValue:@&quot;hobbies&quot; value:@&quot;tennis&quot;];
     &lt;/pre&gt:

     - parameter property: The profile property to add the values for.
     - parameter value: The property value to add to the profile.
     */
    internal func addProfileValue(_ property: String, value: String) {
        if property != "" && value != "" {
            var values = [String]()
            values.append(value)
            self.addProfileValues(property, values: values)
        }
    }

    /**
     Adds property values to the profile. The values from the collection are added to the profile.
     If there are already values for a property the new values will be added.
     Values for a property need to be unique; passing the same values multiple times will have no effect.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     let hobbyArray = [&quot;tennis&quot;, &quot;soccer&quot;]
     client.addProfileValues(&quot;hobbies&quot;, values:hobbyArray)

     // Objective-C:
     NSArray* hobbyArray = [NSArray arrayWithObjects:@&quot;tennis&quot;, @&quot;soccer&quot;, nil];
     [client addProfileValues:@&quot;hobbies&quot; values:hobbyArray];
     &lt;/pre&gt:

     - parameter property: The profile property to add the values for.
     - parameter values: The property values to add to the profile.
     */
    internal func addProfileValues(_ property: String, values: [String]) {
        if property != "" && values.count > 0 {
            self._commitLog.addProperties(property, values: values)
            self._cache.addProperties(property, values: values)
            self._updateRequired = true
        }
    }

    /**
     Sets values on the profile. Passing a property that was already set with values will cause for the old values to be removed.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     client.setProfileValue(&quot;hobbies&quot;, value:&quot;tennis&quot;)

     // Objective-C:
     [client setProfileValue:@&quot;hobbies&quot; value:@&quot;tennis&quot;];
     &lt;/pre&gt:

     - parameter name: The profile property to add the values for.
     - parameter value: The profile value to store.
     */
    internal func setProfileValue(_ name: String, value: String) {
        if name != "" && value != "" {
            var values = [String]()
            values.append(value)
            self.setProfileValues(name, values: values)
        }
    }

    /**
     Sets values on the profile. Passing a property that was already set with values will cause for the old values to be removed.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     let hobbyArray = [&quot;tennis&quot;, &quot;soccer&quot;]
     client.setProfileValues("hobbies", values:hobbyArray)

     // Objective-C:
     NSArray* hobbyArray = [NSArray arrayWithObjects:@&quot;tennis&quot;, @&quot;soccer&quot;, nil];
     [client setProfileValues:@&quot;hobbies&quot; values:hobbyArray];
     &lt;/pre&gt:

     - parameter name: The profile property to add the values for.
     - parameter values: The profile values to store.
     */
    internal func setProfileValues(_ name: String, values: [String]) {
        if name != "" && values.count > 0 {
            self._commitLog.setProperties(name, values: values)
            self._cache.setProperties(name, values: values)
            self._updateRequired = true
        }
    }
    
    /**
     Sets the Mobile Ad ID to the profile.
     Passing a property that was already set with values will cause for the old values to be removed.
     
     Since recent change in IDFA and new rules in iOS 14, BlueConic will no longer automatically call the ASIdentifierManager.
     For an app to capture the Mobile App ID, and where you make the choice to use this service, you can add the Mobile Ad Id with this call

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     guard ASIdentifierManager.shared().isAdvertisingTrackingEnabled else {
         return nil
     }
     client.setMobileAdId(ASIdentifierManager.shared().advertisingIdentifier.uuidString)

     // Objective-C:
     
     
     if([[ASIdentifierManager sharedManager] isAdvertisingTrackingEnabled]){
        NSUUID *idfa = [[ASIdentifierManager sharedManager] advertisingIdentifier];
        [client setMobileAdId: [idfa UUIDString]];
     }
     
     &lt;/pre&gt:

     - parameter mobileAdId: The Mobile Ad ID for that give device (IDFA).
     */
    internal func setMobileAdId(_ mobileAdId: String?) {
        guard let mobileAdId = mobileAdId,
              mobileAdId != "00000000-0000-0000-0000-000000000000" else {
            return
        }
        
        setProfileValue(Constants.ProfileProperties.MOBILE_APP_AD_ID, value: mobileAdId)
        addProfileValue(Constants.ProfileProperties.MOBILE_APP_AD_IDS, value: mobileAdId)
    }

    /**
     Setter for the locale to get the parameters for. By default, the default locale configured in BlueConic is used.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     client.setLocale(&quot;en_US&quot;)

     // Objective-C:
     [client setLocale:@&quot;en_US&quot;];
     &lt;/pre&gt:

     - parameter locale: The locale, e.g. 'en_US'.
     */
    internal func setLocale(_ locale: String) {
        self._locale = locale
    }

    /*
     Returns the session. Default is shared
     :returns: The session
     */
    internal func getSession() -> URLSession {
        
        if let session = self._session {
            return session
        }
        
        let isShared = BlueConicConfiguration.getShared()
        if isShared {
            let session = URLSession.shared
            self._session = session
            return session
        }
        
        BCLog("Using dedicated session and cookie storage")

        let config = URLSessionConfiguration.default
        config.httpCookieStorage = BlueConicCookieStorage()
        config.httpShouldSetCookies = false
        
        let session = URLSession.init(configuration: config)
        self._session = session
        return session
    }
    
    /**
     Setter Enabling BlueConic. By default, BlueConic is enabled.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     client.setEnabled(false)

     // Objective-C:
     [client setEnabled: NO];
     &lt;/pre&gt:

     - parameter isEnabled: To enabled BlueConic.
     */
    internal func setEnabled(_ isEnabled: Bool) {
        self._isEnabled = isEnabled
        BCLog("BlueConic client is \(self._isEnabled ? "enabled": "disabled")")

        self.saveEnabled(self._isEnabled)
    }

    /**
     Checks whether the app was started with simulator data. If so we try to get the username and the the mobile
     session id to connect to the simulator. The intent should look like:
     &quot;&lt;appID&gt;://&lt;hostname&gt;/&lt;username&gt;/&lt;mobilesSessionId&gt;&quot;.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     // Implement in AppDelegate.swift
     func application(application: UIApplication, openURL url: NSURL, sourceApplication: String?, annotation: Any?) -> Bool {
     BlueConicClient.getInstance(nil).setURL(url)
     return true
     }


     // Objective-C:
     // Implement in AppDelegate.m
     - (BOOL)application:(UIApplication *)application openURL:(NSURL *)url sourceApplication:(NSString *)sourceApplication annotation:(id)annotation {
     [[BlueConicClient getInstance:nil] setURL:url];
     return YES;
     }
     &lt;/pre&gt:

     - parameter url: The url retrieved from application.
     */
    internal func setURL(_ url: URL) -> Bool {
        guard let host = url.host,
            let scheme = url.scheme?.lowercased() else {
            BCLog("Unable to connect to the BlueConic Simulator. No host and/or scheme found in url.")
            return false
        }
        
        let urlHost = "https://\(host)"
    
        let packageName = BlueConicConfiguration.getPackageName().lowercased()
        BCLog("Schema data is \(scheme), with package name: \(packageName) via host \(urlHost)")

        if scheme == packageName {
            let params = url.relativePath.components(separatedBy: "/")
            if params.count == 3 {
                
                // Fix: https://jira.blueconic.com/jira/browse/BC-19363
                // Only allow a change in hostname when debug mode is enabled
                //
                // Avoid hijacking in-app opens with a different URL
                //
                // com.test.testapp://non-blueconic-tenant-url/<username>/<mobilesSessionId>"
                // Avoid potential data-leak
                if BlueConicConfiguration.getDebugMode() {
                    // Set host name (Overrule Hostname)
                    self._configuration.setHostName(urlHost)
                }
                
                // Set other simulator data /<username>/<mobileSessionId>
                self._configuration.setSimulatorData(params[1], mobileSessionId: params[2])

                if BlueConicConfiguration.getDebugMode() {
                    // Show an alert view when in debug and connected to the BlueConic Simulator
                    DispatchQueue.main.async {
                        let alert = UIAlertController(title: "Connected to BlueConic Simulator",
                                                      message: "Username: \(params[1]) \nMobile session: \(params[2]) \nHost: \(urlHost)",
                            preferredStyle: UIAlertController.Style.alert)
                        alert.addAction(UIAlertAction(title: "OK", style: UIAlertAction.Style.default, handler: nil))
                        UIApplication.shared.keyWindow?.rootViewController?.present(alert, animated: true, completion: nil)
                    }

                }

                // Sent a request to BlueConic that will notify simulator and reload the profile completely, now that they are both in sync
                if self._configuration.isConnectedToSimulator() {
                    
                    BlueConicClient.synchronized({
                        self.reloadProfile()
                    })
                    
                    BCLog("Connected to the BlueConic Simulator. Username: \(params[1]) Mobile session: \(params[2]) Host: \(urlHost)")
                    return true
                }
            }
        }
        
        return false
    }

    /**
     Setter for the screenname of the current viewcontroller, which can be used in the Simulator as screenname

     - parameter screenName: The screen name of the current viewcontroller'.
     */
    func setScreenName(_ screenName: String) {
        self._screenName = screenName
    }
    
    
    /**
     Getter to retrieve all connections that are included in the page view.

     - returns: A list of BlueConic connections available for the current page view.
     */
    func getConnections() -> [Connection] {
        return self._connections
    }
    
    /**
     Getter to retrieve all segments that the profile has.

     - returns: A list of BlueConic segments where the profile falls in.
     */
    func getSegments() -> [BlueConicSegment] {
        return self._segments
    }
    
    /**
     Check if the profile falls in the segment.

     - parameter segmentId: The ID of the  BlueConic Segment
     - returns: true if the profile falls in the segment based on the ID
     */
    func hasSegment(_ segmentId: String) -> Bool {
        for segment in self._segments {
            if (segment.getId() == segmentId) {
                return true
            }
        }
        return false
    }

    /**
     Set the current context. UIViewController of the app that is displayed on the page.

     - parameter context: The View Controller of the current page
     */
    func setContext(_ context: UIViewController?) {
        if context != nil {
            self._context = context
        }
    }
    
    /**
     Registers an event of the specified type with the given properties. For a &quot;PAGEVIEW&quot; event a screen name can be
     passed, so interactions can be restricted on the where tab in BlueConic.
     For a &quot;VIEW&quot;, &quot;CLICK&quot; or &quot;CONVERSION&quot; event an interactionId should be passed to register the event for.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     client.createEvent(&quot;PAGEVIEW&quot;, properties: [&quot;screenName&quot;: &quot;Main/HOMETAB&quot;])
     client.createEvent(&quot;CLICK&quot;, properties: [&quot;interactionId&quot;: self._context.getInteractionId()])

     // Objective-C:
     [client createEvent:@&quot;PAGEVIEW&quot; properties:@{@&quot;screenName&quot;: @&quot;MAIN/HOMETAB&quot;}];
     [client createEvent:@&quot;CLICK&quot; properties:@{@&quot;interactionId&quot;: [self._context getInteractionId]}];
     &lt;/pre&gt:

     - parameter eventType: The event type. (e.g: &quot;PAGEVIEW&quot;, &quot;VIEW&quot;, &quot;CLICK&quot;, &quot;CONVERSION&quot;)
     - parameter data: A map with properties for the event.

     */
    internal func createEvent(_ eventType: String, data: [String: Any]?, completion: (() -> Void)? = nil) {
        // dialogue interaction events
        if Constants.Event.VIEW == eventType ||
            Constants.Event.CLICK == eventType ||
            Constants.Event.CONVERSION == eventType {
            if let properties = data,
                let interactionId = properties["interactionId"] as? String,
                let event = EventType(rawValue: eventType) {
                self._commitLog.createEvent(event, interactionId: interactionId)
                self._updateRequired = true
            }
            // PAGEVIEW or custom trigger events
        } else {
            if Constants.Event.PAGEVIEW == eventType {
                // Call the onDestroy function for each plugin.
                destroyPlugins()

                if let screenName = self.getScreenNameFromProperties(data) {
                    self.setScreenName(screenName)
                }

                BCLog("Screen name: \(self._screenName)")
            }

            // Handle the Interactions
            self.getInteractions(eventType, context: self.getContext(), screenName: self.getScreenName(), properties: data, completion: completion)
            // return function - handle completion at a later moment.
            return
        }
        
        // handle completion - no need to await on other threads
        DispatchQueue.main.async {
            completion?()
        }
    }
    
    
    /**
     Register plugin class. Assign a custom plugin to the list of available plugins.
     Which are then validated during the PAGEVIEW event, which creates and executes the correct plugin for you.

     &lt;h4&gt;Example&lt;/h4&gt;
     &lt;pre&gt:
     // Swift:
     client.setProfileValue(&quot;hobbies&quot;, value:&quot;tennis&quot;)

     // Objective-C:
     [client setProfileValue:@&quot;hobbies&quot; value:@&quot;tennis&quot;];
     &lt;/pre&gt:

     - parameter pluginClass: The  class of the plugin.
     */
    internal func registerPluginClass(_ pluginClass: AnyClass) {
        let classes: [AnyClass] = [
            pluginClass
        ];

        self.registerPluginClasses(classes)
    }


    /**
     Periodically saves state on disk and on the BlueConic server.
     This method will be invoked periodically by a background thread
     */
    @objc func sync() {
        BlueConicClient.synchronized({
            if !self.isEnabled() {
                return
            }
            
            // Check whether the state of the cache should be written to disk
            if self._cache.checkModified() {
                self.saveCache()
            }

            // Check whether the state of the commitlog should be written to disk
            if self._commitLog.checkModified() {
                self.saveCommitLog()
            }

            // Make sure the commitlog will be flushed if there have been recent changes
            if self._operationQueue.operationCount == 0 && self._updateRequired {
                self.BCLog("Sync started.")
                self.update()
            }
        })
    }

    /**
     Refreshes the local cache
     - parameter completionHandler: The block that should be executed after the operation finished
     */
    func update() {
        self.scheduleUpdate(false)
    }

    func loadProfileValues() {
        self.scheduleUpdate(true)
    }

    /**
     Schedules an update operation the operation queue to flush the commitlog.

     - parameter refreshCache: Indicates whether the local cache should be updated aswell.
     - parameter completionHandler: The block that should be executed after the operation finished
     */
    func scheduleUpdate(_ refreshCache: Bool) {
        if refreshCache {
            self.doRefresh()
        } else {
            self.doUpdate()
        }
    }

    /**
     Flushes the commitlog to the server.
     */
    func doUpdate() {
        self.sendUpdates(false)
    }

    /**
     Flushes the commitlog to the server and updates the local cache.
     */
    func doRefresh() {
        self.sendUpdates(true)
    }
    
    /**
      Update profile
     */
    internal func updateProfile(_ completion: (() -> Void)? = nil) {
        self.sendUpdates(true, completion: completion)
    }

    /**
     Reload profile.
     Always called within another operation.
     */
    func reloadProfile() {
        // Check whether a network connection is available
        if !BlueConicReachability.isConnectedToNetwork() || !self.isEnabled() {
            return
        }

        // Remove all the current connections
        self._connections.removeAll()

        // Keep a list of all method invocations that should be included in the request
        var commands = [BlueConicRequestCommand]()

        // Always send a getProfile call
        commands.append(BlueConicConnector.getProfileCommand(CallId.profile.rawValue))

        // Send a getProperties call
        commands.append(BlueConicConnector.getGetPropertiesCommand(CallId.getProperties.rawValue, hash: nil))

        if self._configuration.isConnectedToSimulator() {
            self.BCLog("Connected with simulator")
            // in the simulator status also throw another pageview event for the reloaded profile to make sure it is using the new profile
            commands.append(BlueConicConnector.getInteractionCommand(CallId.interactions.rawValue, eventType: Constants.Event.PAGEVIEW))
        }

        let connector = BlueConicConnector(session: self.getSession(), userName: self._configuration.getSimulatorUsername(), mobileSessionId: self._configuration.getSimulatorSessionId())
        let domainGroupUrl = BlueConicConfiguration.getDomaingGroupUrl(self._configuration.getHostName(), domainGroup: self._cache.getDomainGroup(), zoneId: self.getZoneId())

        let result = connector.execute(domainGroupUrl, calls: commands, screenName: self._configuration.isConnectedToSimulator() ? self._screenName : nil)

        guard let response = result.response else {
            // Something went wrong, see the error what happend.
            if let error = result.error {
                self.BCLog("Unable to reload the profile from BlueConic, reason: \(error.localizedDescription)", flag: Constants.Flag.ERROR)
            } else {
                self.BCLog("Unable to reload the profile from BlueConic", flag: Constants.Flag.WARN)
            }
            return
        }

        let profileResponse = response["\(CallId.profile.rawValue)"]
        let propertiesResponse = response["\(CallId.getProperties.rawValue)"]

        // Process the results no need to reload another time.
        let _ = self.processProfileResult(profileResponse)
        self.processPropertiesResult(propertiesResponse, isUpdate: false)

        self.BCLog("Done reloading")
    }
    
    /**
     Flushes the commitlog to the server.

     - parameter refreshCache: Indicates whether the local cache should be updated aswell.
     - returns: true if the operation completed successfully, false otherwise.
     */
    func sendUpdates(_ refreshCache: Bool) {
        self.sendUpdates(refreshCache, completion: nil)
    }
    
    /**
     Flushes the commitlog to the server.

     - parameter refreshCache: Indicates whether the local cache should be updated aswell.
     - parameter completion: completion handler.
     - returns: true if the operation completed successfully, false otherwise.
     */
    func sendUpdates(_ refreshCache: Bool, completion: (() -> Void)? = nil) {
        // Check whether a network connection is available
        if !BlueConicReachability.isConnectedToNetwork() || !self.isEnabled() {
            BCLog("Sync failed! Unable to connect to a network", flag: Constants.Flag.WARN)
            
            // handle completion - no need to await on other threads
            DispatchQueue.main.async {
                completion?()
            }
            return
        }
        
        // Execute all required methods an check whether an error occurred
        var domainGroupUrl: String?
        var connector: BlueConicConnector?
        var commands = [BlueConicRequestCommand]()
        
        self._operationQueue.addOperation({
            BlueConicClient.synchronized({
                // Reset the update required flag. Try submitting change only once in case of an active connection
                self._updateRequired = false

                // Merge and clear the commitlog with the request commitlog
                self._requestCommitLog.mergeCommitLog(self._commitLog)

                // Get the properties that should be added
                let toAdd: [String: BlueConicCommitEntry] = self._requestCommitLog.getProperties(OperationType.ADD)

                // Get the properties that should be set
                let toSet: [String: BlueConicCommitEntry] = self._requestCommitLog.getProperties(OperationType.SET)

                // Get the events that should be created
                let toEvents: [BlueConicCommitEntry] = self._requestCommitLog.getEvents()

                // Check whether there is work to do
                if toAdd.count == 0 && toSet.count == 0 && toEvents.count == 0 && !refreshCache {
                    self.BCLog("Sync successful.")
                    
                    // handle completion - no need to await on other threads
                    DispatchQueue.main.async {
                        completion?()
                    }
                    return
                }

                // Keep a list of all method invocations that should be included in the request
                commands = BlueConicConnector.setUpdateCommand(toAdd, setProperties: toSet, events: toEvents)

                // Add a call if the local cache of profile properties should be updated
                if refreshCache {
                    // Get the hash of the properties that should be get
                    let hash: String = "\(self._cache.getHash())"

                    commands.append(BlueConicConnector.getGetPropertiesCommand(CallId.getProperties.rawValue, hash: hash))
                }

                // Get BlueConicConnector instance
                connector = BlueConicConnector(session: self.getSession(), userName: self._configuration.getSimulatorUsername(), mobileSessionId: self._configuration.getSimulatorSessionId())
                domainGroupUrl = BlueConicConfiguration.getDomaingGroupUrl(self._configuration.getHostName(), domainGroup: self._cache.getDomainGroup(), zoneId: self.getZoneId())

                self.BCLog("Domain group url: \(domainGroupUrl ?? "")")
            })
            
           guard let connector = connector,
               let domainGroupUrl = domainGroupUrl else {
               self.BCLog("Unable to get the interactions from BlueConic, reason: No connector or commands established", flag: Constants.Flag.ERROR)
               return
           }

            let result = connector.execute(domainGroupUrl, calls: commands, screenName: nil)
            BlueConicClient.synchronized({
                if result.error != nil || result.response == nil {
                    // Something went wrong, see the error what happend.
                    if let error = result.error {
                        self.BCLog("Unable to get the interactions from BlueConic, reason: \(error.localizedDescription)", flag: Constants.Flag.ERROR)
                    } else {
                        self.BCLog("Unable to get the interactions from BlueConic", flag: Constants.Flag.WARN)
                    }

                    self.BCLog("Sync failed", flag: Constants.Flag.WARN)
                    
                    // handle completion - no need to await on other threads
                    DispatchQueue.main.async {
                        completion?()
                    }
                    return
                }

                // Process the result of the getProfilecall
                let reload = self.processProfileResult(result.response!["\(CallId.profile.rawValue)"])

                // Process the result of the getProperties call to update the cache
                self.processPropertiesResult(result.response!["\(CallId.getProperties.rawValue)"], isUpdate: false)

                // The changes have been committed so the commit can be rest
                self._requestCommitLog.clearAll()

                // Reload the profile is there is a profile switch going on.
                if reload {
                    self.reloadProfile()
                }
                // The operation has been executed successfully
                self.BCLog("Sync succesful.")
                
                
                // handle completion - no need to await on other threads
                DispatchQueue.main.async {
                    completion?()
                }
                
                return
            })
        })
    }
    
    /*
     Get the context object
     :returns: The contxt object, also could return <code>nil</code>
     */
    private func getContext() -> UIViewController? {
        return self._context
    }
    
    /*
     Get the currnet locale
     :returns: The locale, also could return <code>nil</code>
     */
    private func getLocale() -> String? {
        return self._locale
    }

    /*
     Get the plugin by class name.

     :param: ClassName with project name, e.g. BlueConicPlugins.Banner.
     :returns: The which can be initialized, also could return <code>nil</code>.
     */
    // Plugin class became @deprecated use BlueConicPlugin
    func getPlugin(_ className: String) -> AnyClass? {
        // Get the Plugin Type from the BlueConicClient
        if let type: AnyClass = self.getPluginClass(className) {
            return type
        }

        // Get the Plugin Type from the current project (source).
        if let info = Bundle.main.infoDictionary,
            let projectName: String = info["CFBundleExecutable"] as? String,
            let index = className.range(of: ".")?.upperBound {

            // it's allowed to have spaces in a Project name, replace them with a underscore
            let replacedProjectName = projectName.components(separatedBy: " ").joined(separator: "_")
            
            // gather the customers project name to check if the plugin contains in project
            let pluginName = String(className[index...])
            let pluginClassName = replacedProjectName + "." + pluginName

            if let customType: AnyClass = self.getPluginClass(pluginClassName) {
                return customType
            }
        } else {
            BCLog("Project name cannot be found or plugins class <\(className)> is invalid", flag: Constants.Flag.WARN)
        }
        // Cache negative results too
        self._plugins[className] = nil
        return nil
    }

    /*
     Get the plugin class, try to get if based on the swift and objective-c style to retrieve a class.
     Swift needs a project name and plugin name to retrieve a class.
     Objective-c only needs the plugin name.

     :param: ClassName with project name, e.g. BlueConicPlugins.Banner.
     :returns: The Plugin.Type which can be initialized, also could return <code>nil</code>.
     */
    // Plugin class became @deprecated use BlueConicPlugin
    func getPluginClass(_ className: String) -> AnyClass? {
        var pluginClass: AnyClass? = nil
        // Swift needs ProjectName.PluginName
        if let swiftPluginClass = self.getCachedPluginClass(className) {
            pluginClass = swiftPluginClass
            // Objective-C only needs PluginName
        } else if let index = className.range(of: ".")?.upperBound {
            let pluginName = String(className[index...])
            if let objectivecPluginClass = self.getCachedPluginClass(pluginName) {
                pluginClass =  objectivecPluginClass
            }
        }

        return pluginClass
    }

    /*
     Register all out of the box plugin classes
     */
    func registerPluginClasses() {
        let classes: [AnyClass] = [
            BehaviorListener.self,
            EngagementRankingListener.self,
            EngagementScoreListener.self,
            FormListener.self,
            FullscreenLightbox.self,
            PreferredHourListener.self,
            GlobalListener.self
        ];

        self.registerPluginClasses(classes)
        BCLog("Following plugins are available: \(self._plugins.keys.joined(separator: ", "))")
    }
    
    /*
     Store all Plugin classes in cache, which can be used to hardcode set a plugin class.
     */
    func registerPluginClasses(_ classes: [AnyClass]) {
        for pluginClass in classes {
            let className = NSStringFromClass(pluginClass)
            self.registerPluginClass(pluginClass, className: className)
        }
    }
    
    /*
     Register plugin class
     */
    internal func registerPluginClass(_ pluginClass: AnyClass, className: String) {
        if let _ = pluginClass as? BlueConicPlugin.Type {
            // Don't update the plugin if it is already in cache
            if self._plugins[className] != nil {
                return
            }

            // Add all plugins based on className and also on BlueConicClient.className
            self._plugins[className] = pluginClass
            if let pluginClassName: String = className.components(separatedBy: ".").last,
                self._plugins["BlueConicClient.\(pluginClassName)"] == nil {
                self._plugins["BlueConicClient.\(pluginClassName)"] = pluginClass
            }
            BCLog("The plugin '\(className)' is registered.")
        } else {
            BCLog("The class '\(className)' is not of type 'BlueConicPlugin' and will not be registered.")
        }
    }
    
    

    /*
     Get the plugin by class name and add cache the results

     :param: ClassName with project name, e.g. BlueConicPlugins.Banner.
     :returns: The Plugin.Type which can be initialized, also could return <code>nil</code>.
     */
    // Plugin class became @deprecated use BlueConicPlugin
    func getCachedPluginClass(_ className: String) -> AnyClass? {
        // Check if there is a cached plugin
        if let cachedPlugin: AnyClass = self._plugins[className] {
            return cachedPlugin
        }
        // FIXME: NSClassFromString in an Objective-C type
        if let pluginClass: AnyClass = NSClassFromString(className) {
            BCLog("Plugin found at <\(className)>")

            // Cache the results
            self._plugins[className] = pluginClass
            return pluginClass
        }
        return nil
    }

    /**
     Private method. Retrieves interactions.

     - parameter context: The active ViewController of the app.
     - parameter screenName: the screen name in the app (e.g. &quot;Home/MyPage&quot;)
     */
    func getInteractions(_ eventType: String, context: UIViewController?, screenName: String, properties: [String: Any]?, completion: (() -> Void)? = nil) {
        // Execute all required methods an check whether an error occurred
        var domainGroupUrl: String?
        var connector: BlueConicConnector?
        var commands = [BlueConicRequestCommand]()

        let isPageview: Bool = Constants.Event.PAGEVIEW == eventType;
        
        self._operationQueue.addOperation({
            BlueConicClient.synchronized({
                let eventManager = BlueConicEventFactory.getInstance()

                // Check whether a network connection is available
                if !BlueConicReachability.isConnectedToNetwork() || !self.isEnabled() {
                    DispatchQueue.main.async {
                        completion?()
                    }
                    return
                }

                
                if isPageview {
                    // Reset the update required flag. Try submitting change only once in case of an active connection
                    self._updateRequired = false

                    // Merge and clear the commitlog with the request commitlog
                    self._requestCommitLog.mergeCommitLog(self._commitLog)

                    // Get the properties that should be added
                    let toAdd: [String: BlueConicCommitEntry] = self._requestCommitLog.getProperties(OperationType.ADD)

                    // Get the properties that should be set
                    let toSet: [String: BlueConicCommitEntry] = self._requestCommitLog.getProperties(OperationType.SET)

                    // Get the events that should be created
                    let toEvents: [BlueConicCommitEntry] = self._requestCommitLog.getEvents()

                    // Get the hash of the properties that should be get
                    let hash: String = "\(self._cache.getHash())"

                    // Remove all the current connections
                    self._connections.removeAll()
                    
                    // clear all events for the EventManager
                    eventManager.clearEvents();

                    // Keep a list of all method invocations that should be included in the request
                    commands = BlueConicConnector.setUpdateCommand(toAdd, setProperties: toSet, events: toEvents)

                    // Send a getProperties call
                    commands.append(BlueConicConnector.getGetPropertiesCommand(CallId.getProperties.rawValue, hash: hash))
                } else {
                    // This is a custom event, also register as AdvancedEvent, when it has context
                    if let eventContext = properties?["context"] as? [String] {
                        eventManager.publish(AdvancedEvent(eventName: eventType, context: eventContext))
                    } else if let eventValue = properties?["context"] as? String {
                        eventManager.publish(AdvancedEvent(eventName: eventType, context: [eventValue]))
                    }
                }

                // Send a getInteraction call
                commands.append(BlueConicConnector.getInteractionCommand(CallId.interactions.rawValue, eventType: eventType, properties: properties))

                connector = BlueConicConnector(session: self.getSession(), userName: self._configuration.getSimulatorUsername(), mobileSessionId: self._configuration.getSimulatorSessionId())
                domainGroupUrl = BlueConicConfiguration.getDomaingGroupUrl(self._configuration.getHostName(), domainGroup: self._cache.getDomainGroup(), zoneId: self.getZoneId())
                self.BCLog("Domaingroup URL: \(domainGroupUrl ?? "")")

            })
            
            guard let connector = connector,
                let domainGroupUrl = domainGroupUrl else {
                self.BCLog("Unable to get the interactions from BlueConic, reason: No connector or commands established", flag: Constants.Flag.ERROR)
                DispatchQueue.main.async {
                   completion?()
                }
                return
            }
            let result = connector.execute(domainGroupUrl, calls: commands, screenName: screenName)
                
            BlueConicClient.synchronized({
                guard let response = result.response else {
                    // Something went wrong, see the error what happend.
                    if let error = result.error {
                        self.BCLog("Unable to get the interactions from BlueConic, reason: \(error.localizedDescription)", flag: Constants.Flag.ERROR)
                    } else {
                        self.BCLog("Unable to get interactions from BlueConic", flag: Constants.Flag.WARN)
                    }
                    
                    DispatchQueue.main.async {
                        completion?()
                    }
                    return
                }

                let profileResponse = response["\(CallId.profile.rawValue)"]
                let propertiesResponse = response["\(CallId.getProperties.rawValue)"]
                let interactionsResponse = response["\(CallId.interactions.rawValue)"]
                var reload = false

                if isPageview {
                    // Process the results
                    reload = self.processProfileResult(profileResponse)
                    self.processPropertiesResult(propertiesResponse, isUpdate: true)
                    self.processConnectionsResult(interactionsResponse)
                    self.processSegmentsResult(interactionsResponse)
                }

                self.processInteractionsResult(context, resultObject: interactionsResponse, isPageview: isPageview)

                if isPageview {
                    // The changes have been committed so the commit can be rest
                    self._requestCommitLog.clearAll()

                    if reload {
                        self.reloadProfile();
                    }
                }
                
                DispatchQueue.main.async {
                    completion?()
                }
            })
        })
    }

    /**
     Update the profile ID.
     Set and save the new given profile ID, and clear the old cache if the profile changed.

     - parameter newProfileId: The new profile ID
     */
    private func _updateProfileId(_ newProfileId: String?) -> Bool {
        var changed = false
        if self.hasToUpdateProperty(newProfileId, storedProperty: self._profileId) {
            changed = self._profileId != nil

            // Clear the cache if the profile id changed
            if changed {
                BCLog("New profile ID: \(newProfileId ?? "") - Old profile ID: \(self._profileId ?? "")")
                self._cache.clear()
            }

            // Store the new profile id
            self._profileId = newProfileId

            self.saveProfileId(self._profileId!)
        }
        return changed
    }

    /**
     Update the Zone ID
     Set and save the new given Zone ID, and clear the old cache if the profile changed.
     When the Zone ID is set, it will add it to each server request, so the load balancer known which server to use.

     - parameter newZoneId: The new zone ID
     */
    private func _updateZoneId(_ newZoneId: String?) {
        if self.hasToUpdateProperty(newZoneId, storedProperty: self.getZoneId()) {
            self._zoneId = newZoneId
        }
    }
    
    /*
     Load profile id from file

     :returns: profileId or empty string
     */
    func loadProfileId() -> String {
        return BlueConicDataManager.loadStringFromFile(fileName: Constants.Files.ID) ?? ""
    }
    
    /*
     Load enabled from file

     :returns: enabled or true
     */
    func loadEnabled() -> Bool {
        let data = BlueConicDataManager.loadStringFromFile(fileName: Constants.Files.ENABLED) ?? "true"
        return data == "false" ? false : true
    }
    
    /**
     Write the profileId to a specified file
     - parameter profileId: The profile identifier retrieved from the server
     */
    func saveEnabled(_ isEnabled: Bool) {
        let data = "\(isEnabled)"
        BlueConicDataManager.writeStringToFile(data: data, fileName: Constants.Files.ENABLED)
    }

    /**
     Write the profileId to a specified file
     - parameter profileId: The profile identifier retrieved from the server
     */
    func saveProfileId(_ profileId: String) {
        BlueConicDataManager.writeStringToFile(data: profileId, fileName: Constants.Files.ID)
    }

    /**
     Load the latest saved BlueConicCommitLog
     - parameter fileLocation: The location of the stored commit log file.
     - returns: a BlueConicCommit log, could return a new instance of BlueConicCommitLog if the file didn't exists
     */
    func loadCommitLog(_ fileLocation: String) -> BlueConicCommitLog {
        return BlueConicDataManager.loadFromFile(fileName: fileLocation, type: BlueConicCommitLog.self) ?? BlueConicCommitLog()
    }

    /**
     Save the current BlueConicCommitLog to a specified file
     If the file already exists it will be overwritten.
     */
    func saveCommitLog() {
        BlueConicDataManager.writeToFile(self._commitLog, fileName: Constants.Files.COMMITLOG)
        BlueConicDataManager.writeToFile(self._requestCommitLog, fileName: Constants.Files.REQUEST_COMMITLOG)
    }

    /**
     Load the latest saved BlueConicCache
     - returns: a BlueConicCache log, could return a new instance of BlueConicCache if the file didn't exists
     */
    func loadCache() -> BlueConicCache {
        return BlueConicDataManager.loadFromFile(fileName: Constants.Files.CACHE, type: BlueConicCache.self) ?? BlueConicCache()
    }

    /**
     Save the current BlueConicCache to a specified file
     If the file already exists it will be overwritten.
     */
    func saveCache() {
        BlueConicDataManager.writeToFile(self._cache, fileName: Constants.Files.CACHE)
    }

    /*
     Returns the Zone ID
     - returns: zoneId or nil, when there isn't a zoneId set.
     */
    func getZoneId() -> String? {
        return self._zoneId
    }

    /**
     Returns whether a property is valid to update.

     - retruns true when it is valid to update the property.
     */
    func hasToUpdateProperty(_ property: String?, storedProperty: String?) -> Bool {
        // Check wether the property is new and valid
        return property != nil && property != storedProperty
    }

    /**
     Process the result of the getProperties call
     Check if the interaction type is registered as plugin
     Create a new instance of the plugins that are equal to the interaction type and add it to the array

     - parameter interactions: All interactions from the method call response.
     - returns: plugins All interactions that are have an interaction type that is equal to a registered plugin
     */
    // Plugin class became @deprecated use BlueConicPlugin
    func getRegisteredPluginInteractions(_ interactions: [[String: Any]]) -> [AnyObject] {
        var pluginsArray = [AnyObject]()

        for interaction in interactions {
            if let interactionId = interaction[Constants.Interactions.ID] as? String,
                let interactionPluginClass: String = interaction[Constants.Interactions.CLASS] as? String,
                let interactionType = interaction[Constants.Interactions.TYPE] as? String,
                let pluginClass: AnyClass = getPlugin(interactionPluginClass) {
                BCLog("Plugin class <\(interactionPluginClass)> found! For the id: \(interactionType) and interaction: \(interactionId)")

                let interactionContext = InteractionContext(interaction: interaction, context: self.getContext(), locale: self.getLocale())
                if let pluginClass = pluginClass as? BlueConicPlugin.Type {
                    let pluginItem: AnyObject = pluginClass.init(client: BlueConicClient.getInstance(self.getContext()), context: interactionContext)
                    pluginsArray.append(pluginItem)
                }
            } else {
                BCLog("Plugin class <\(interaction[Constants.Interactions.CLASS] as? String ?? "")> not found for type id: \(interaction[Constants.Interactions.TYPE] as? String ?? ""); and interaction: \(interaction[Constants.Interactions.ID] as? String ?? "")", flag: Constants.Flag.WARN)
            }
        }
        return pluginsArray
    }

    /**
     Check if the data contains connections and append them the list.
     It changes the parameters to a dictionary, so it is equals to Android and easier to access.
     - parameter connections: All the connections data from the method call response.
     - returns: List of connections All the connections that are create by the data
     */
    func addConnections(_ connections: [[String: Any]]) -> [Connection] {
        var connectionsArray = [Connection]()
        for connection in connections {
            // Only get the needed parameters. use the first locale that is available.
            if let connectionId = connection[Constants.Connections.ID] as? String,
                let connectionParameters = connection[Constants.Connections.PARAMETERS] as? [[String: Any]],
                let parameters = connectionParameters[0][Constants.Connections.PARAMETER] as? [[String: Any]] {

                // Change the parameters to an useable dictionary.
                let connectionDictinoary = BlueConicConfiguration.parametersToDictionary(parameters)

                // Add a new connection to the list.
                connectionsArray.append(Connection(id: connectionId, parameters: connectionDictinoary))
            }
        }
        BCLog(String(connectionsArray.count) + " connection(s) added!")
        return connectionsArray
    }
    
    /**
     Check if the data contains segments and append them the list.
     It changes the parameters to a dictionary, so it is equals to Android and easier to access.
     - parameter segments: All the segments data from the method call response.
     - returns: List of segments All the segments that are create by the data
     */
    func addSegments(_ segments: [[String: Any]]) -> [BlueConicSegment] {
        var segmentResult = [BlueConicSegment]()
        for segment in segments {
            // Only get the needed parameters. use the first locale that is available.
            if let segmentId = segment[Constants.Segments.ID] as? String,
                let segmentName = segment[Constants.Segments.NAME] as? String {

                // Add a new connection to the list.
                segmentResult.append(BlueConicSegment(id: segmentId, name: segmentName))
            }
        }
        BCLog(String(segmentResult.count) + " segment(s) added!")
        return segmentResult
    }

    /**
     Calls the onLoad() method of each plugin

     - parameter context: The active ViewController
     - parameter plugins: The created plugins from the response
     - parameter isPageview: if it is a real pageview remove all cusrrent plugins
     */
    func loadPlugins(_ plugins: [AnyObject], isPageview: Bool) {
        // Run the following on the Main Thread
        if plugins.count > 0 {

            if isPageview {
                self._currentPlugins.removeAll()
            }

            for plugin in plugins {
                DispatchQueue.main.async {
                    plugin.onLoad()
                }
                self._currentPlugins.append(plugin)
            }
            DispatchQueue.main.async {
                let eventManager = BlueConicEventFactory.getInstance()
                eventManager.handleAllEvents();
            }
        }
    }

    /**
     Calls the onDestroy() method of each plugin.
     */
    func destroyPlugins() {
        for plugin in self._currentPlugins {
            DispatchQueue.main.async {
                // onDestroy is an optional function, check before calling
                plugin.onDestroy?()
            }
        }
    }

    /**
     Store the domain groups.
     - parameter domainGroup: the domain group to set.
     */
    func setDomainGroup(_ domainGroup: String) {
        self._cache.setDomainGroup(domainGroup)
    }

    /*
     Get the domain group.
     */
    func getDomainGroup() -> String? {
        return self._cache.getDomainGroup()
    }


    /*
     Internal debug logging
     - parameter message:
     */
    func BCLog(_ message: String, flag: String = Constants.Flag.INFO) {
        BlueConicConfiguration.log(message, flag: flag)
    }
    /**
     Process the result of the getProfile call
     Checks if the new given profile id is equals the current, set the new profile id as current if they are not equal

     - parameter resultObject: The response from the method call of type Dictionary
     */
    func processProfileResult(_ resultObject: Any?) -> Bool {
        // Process the result of the getProfilecall

        // Get the dictionary of the result from the getProfile call
        if let profileProperties = resultObject as? [String: Any] {

            // Get the profile id returned by the server
            let newProfileId: String? = profileProperties[Constants.ProfileProperties.ID] as? String

            // Get the profile id returned by the server
            let newZoneId: String? = profileProperties[Constants.ProfileProperties.ZONEID] as? String

            // Get the domainGroupId from result and cache the domain group.
            if let resultDictionary = profileProperties[Constants.RESULT] as? [String: Any],
                let domainGroupId = resultDictionary[Constants.DOMAINGROUPID] as? String {
                setDomainGroup(domainGroupId)
            }


            // Check whether the zone id is valid
            self._updateZoneId(newZoneId)

            // Check whether the id is valid
            return self._updateProfileId(newProfileId)
        }
        return false
    }

    /**
     Process the result of the getProperties call from the getProperties call
     Clear the cache and set the new properties if the response is not empty

     - parameter resultObject: The response from the method call of type Dictionary
     - parameter isUpdate: whether to update the profile
     */
    func processPropertiesResult(_ resultObject: Any?, isUpdate: Bool) {
        // Get the dictionary of the result from the getProperties call
        if let resultObject = resultObject as? [String: Any],
            let results = resultObject[Constants.RESULT] as? [String: Any],
            let properties = results[Constants.PROPERTIES] as? [String: [String]] {

            if !isUpdate {
                self._cache.clear()
            } else {
                BCLog("Updated properties: \(properties)")
            }

            self._cache.setProperties(properties)
        }
    }

    /**
     Process the result of the getInteractions call
     Check if any of the interactions is registered and if so then call the onLoad() method of that specified plugin.
     - parameter context: The current view controller.
     - parameter resultObject: The response from the method call of type Dictionary
     - parameter isPageview: reload loading when it is a real pageview
     */
    func processInteractionsResult(_ context: UIViewController?, resultObject: Any?, isPageview: Bool) {
        if let interactionDictionary = resultObject as? [String: Any],
            let resultDictionary = interactionDictionary[Constants.RESULT] as? [String: Any],
            let interactions = resultDictionary[Constants.Interactions.INTERACTIONS_KEY] as? [[String: Any]] {
            let pluginsArray: [AnyObject] = getRegisteredPluginInteractions(interactions)
            // Call the onLoad function of each plugin that is registered and retrieved from the server.
            loadPlugins(pluginsArray, isPageview: isPageview)
        }
    }

    /**
     Process the result of the getInteractions call
     Add all the received connections to the connections list.

     - parameter resultObject: The response from the method call of type Dictionary
     */
    func processConnectionsResult(_ resultObject: Any?) {
        if let interactionDictionary = resultObject as? [String: Any],
            let resultDictionary = interactionDictionary[Constants.RESULT] as? [String: Any],
            let connections = resultDictionary[Constants.Connections.CONNECTIONS_KEY] as? [[String: Any]] {
            self._connections = self.addConnections(connections)
        }
    }
    
    /**
     Process the result of the getInteractions call
     Add all the received segments to the segments list.

     - parameter resultObject: The response from the method call of type Dictionary
     */
    func processSegmentsResult(_ resultObject: Any?) {
        if let interactionDictionary = resultObject as? [String: Any],
            let resultDictionary = interactionDictionary[Constants.RESULT] as? [String: Any],
            let segments = resultDictionary[Constants.Segments.SEGMENTS_KEY] as? [[String: Any]] {
            self._segments = self.addSegments(segments)
        }
    }
}
